{
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PLATFORM  : Windows 9x/NT/XP - Delphi 5,6,7
PROJECT   : eXpress Application Manager
FILE      : U_XAM_Debug.pas
PURPOSE   : Exception manager.
            Use the IXAMDebug interface to access the singleton debug object.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
LICENSE   : LGPL
AUTHOR    : Copyright 2003 Olivier Rogier 
WEBSITE   : http://www.ordisoftware.info
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CREATION  : v 1.0.0 - 01/2003
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}

unit U_XAM_Debug;

{
#######################################################################################################################
                                                       INTERFACE
#######################################################################################################################
}

interface

uses
  Classes, SysUtils;

type
  // Action type to do on exception protected by enter/leave system
  // Rem : an EAbort exception always break the execution
  TXAMExceptionAction  = ( ExceptMask, ExceptRaise, ExceptStop );

  // Events for doing actions on before & after an exception
  TXAMShowExceptionEvent   = procedure (Sender : TObject; E : Exception) of object;
  TXAMBeforeExceptionEvent = procedure (Sender : TObject; E : Exception; var Process : boolean) of object;
  TXAMOnExceptionEvent     = procedure (Sender : TObject; E : Exception) of object;
  TXAMAfterExceptionEvent  = procedure (Sender : TObject; E : Exception; Processed : boolean) of object;

  // Exception details
  TXAMExceptionInfos = record
    E           : Exception;    // Exception object
    EClassName  : string;       // Exception class name
    EInherits   : string;       // Exception parents class names
    EFullName   : string;       // Exception full name
    EMessage    : string;       // Exception Message
    EAddress    : string;       // Exception address in hexadecimal
    AppName     : string;       // Application name
    AppPath     : string;       // Application path
    AppVer      : string;       // Application version
    Date        : string;       // Exception date
    User        : string;       // User session name
    Computer    : string;       // Computer name
    CodeUnit    : string;       // Pascal unit name
    CodeLine    : string;       // Pascal unit line
    CodeClass   : string;       // Code class name
    CodeProc    : string;       // Procedure name
    SenderClass : string;       // Sender class name
    SenderName  : string;       // Sender component name
    CallStack   : string;       // Calls stack
  end;

  // **********************************************************************************************
  // TXAMMapFile
  // **********************************************************************************************
  // PURPOSE : Link map file manager
  // **********************************************************************************************
  TXAMMapFile = class
  private
    FStackInfos : record
                    LastEBP    : pointer;
                    CallerAddr : pointer;
                  end;
    FFileName   : TFileName;        // Map file name
    FPresent    : boolean;          // True if map file loaded
    FSegments   : TStrings;         // Segments List
    FAdress     : TStrings;         // Addresses List
    FLines      : TStrings;         // Lines List
    FAddr       : cardinal;         // Current Addresse for infos below
    FUnitName   : string;           // Unit name
    FProcName   : string;           // Proc name
    FProcAddr   : integer;          // Proc address
    FLineNo     : string;           // Line number
  private
    procedure UpdateUnitName;
    procedure UpdateProcName;
    procedure UpdateLineNo;
    procedure SetFileName(Value : TFileName);
    procedure SetAddr(Value : cardinal);
  public
    constructor Create;
    destructor  Destroy; override;
    function    FirstCallStack : boolean;
    function    NextCallStack : boolean;
  published
    property FileName : TFileName read FFileName  write SetFileName;
    property Addr     : cardinal  read FAddr      write SetAddr;
    property Present  : boolean   read FPresent;
    property UnitName : string    read FUnitName;
    property ProcName : string    read FProcName;
    property ProcAddr : integer   read FProcAddr;
    property LineNo   : string    read FLineNo;
  end;

  // **********************************************************************************************
  // IDebug
  // **********************************************************************************************
  // PURPOSE : Exception manager interface
  // **********************************************************************************************
  IXAMDebug = interface
    function  GetMapFile : TXAMMapFile;
    function  GetCallStack(SkipFromLast : cardinal = 0; GetOne : boolean = false) : string;
    function  IsActive : boolean;
    procedure SetActive(Value : boolean);
    procedure SetShowException(Value : TXAMShowExceptionEvent);
    function  GetShowException : TXAMShowExceptionEvent;
    procedure SetBeforeException(Value : TXAMBeforeExceptionEvent);
    function  GetBeforeException : TXAMBeforeExceptionEvent;
    procedure SetOnException(Value : TXAMOnExceptionEvent);
    function  GetOnException : TXAMOnExceptionEvent;
    procedure SetAfterException(Value : TXAMAfterExceptionEvent);
    function  GetAfterException : TXAMAfterExceptionEvent;
    function  GetExceptionInfos(Sender : TObject; E : Exception) : TXAMExceptionInfos;
    procedure ProcessException(Sender : TObject; E : Exception; DoShow : boolean);
    procedure EnterProc(ActionType : TXAMExceptionAction = ExceptRaise; DoShow : boolean = false); overload;
    procedure EnterProc(Params : array of const; ActionType : TXAMExceptionAction = ExceptRaise; DoShow : boolean = false); overload;
    procedure LeaveProc;
    function  DoProcess : boolean;
  end;

function  XAMDebugger : IXAMDebug;
procedure XAMCheckIndex(IndexName : string; IndexValue, MinValue, MaxValue : integer; ExceptClass : ExceptClass); overload;
procedure XAMCheckIndex(IndexName : string; IndexValue, MinValue, MaxValue : int64;   ExceptClass : ExceptClass); overload;

{
#######################################################################################################################
                                                     IMPLEMENTATION
#######################################################################################################################
}

implementation

uses
  Windows, Forms, Dialogs, U_XAM_Global, U_XAM_System, F_XAM_Exception, U_XAM_LogFile;

const
  ERR_INVALID_MAP_FILE   = 'Invalid Map File : %s.';
  ERR_MAP_FILE_NOT_FOUND = 'Map File not found : %s.';
  MAPFILE_EXT            = '.map';
  MAPFILE_M_1            = 7;
  MAPFILE_M_2            = 16;
  MAPFILE_FCT            = 22;
  MAPFILE_LINE_1         = 1;
  MAPFILE_LINE_2         = 13;
  MAPFILE_LINE_3         = 20;
  ADDR_LENGTH            = 8;

type
  // procedure informations for enter/leave system
  TCallProcInfo = class(TCollectionItem)
  public
    Caption : string;
    Action  : TXAMExceptionAction;
    Show    : boolean;
  end;

  // Exceptions
  EMapFileException = class(Exception);   // Errors from MapFile manipulations
  EDebugException   = class(Exception);   // Errors from Debug manipulations

  // **********************************************************************************************
  // TDebug
  // **********************************************************************************************
  // PURPOSE : Exception Manager
  // **********************************************************************************************
  TXAMDebug = class(TObject, IXAMDebug)
  private
    FOldHandler      : TExceptionEvent;         // Previous exception manager
    FObjectInfos     : TXAMObjectInfos;
    FProcCalls       : TCollection;             // Enter/Leave proc calls list
    FProcCallsMargin : integer;
    FMapFile         : TXAMMapFile;
    FActive          : boolean;
    FShowException   : TXAMShowExceptionEvent;
    FBeforeException : TXAMBeforeExceptionEvent;
    FOnException     : TXAMOnExceptionEvent;
    FAfterException  : TXAMAfterExceptionEvent;
  private
    function  QueryInterface(const IID: TGUID; out Obj): Integer; stdcall;
    function  _AddRef: Integer; stdcall;
    function  _Release: Integer; stdcall;
    procedure HandleException(Sender : TObject; E : Exception);
    procedure ShowException(Sender : TObject; E : Exception);
  public
    constructor Create;
    destructor  Destroy; override;
    function    IsActive : boolean;
    procedure   SetActive(Value : boolean);
    function    GetMapFile : TXAMMapFile;
    function    GetCallStack(SkipFromLast : cardinal = 0; GetOne : boolean = false) : string;
    procedure   SetShowException(Value : TXAMShowExceptionEvent);
    function    GetShowException : TXAMShowExceptionEvent;
    procedure   SetBeforeException(Value : TXAMBeforeExceptionEvent);
    function    GetBeforeException : TXAMBeforeExceptionEvent;
    procedure   SetOnException(Value : TXAMOnExceptionEvent);
    function    GetOnException : TXAMOnExceptionEvent;
    procedure   SetAfterException(Value : TXAMAfterExceptionEvent);
    function    GetAfterException : TXAMAfterExceptionEvent;
    function    GetExceptionInfos(Sender : TObject; E : Exception) : TXAMExceptionInfos;
    procedure   ProcessException(Sender : TObject; E : Exception; DoShow : boolean);
    procedure   EnterProc(ActionType : TXAMExceptionAction = ExceptRaise; DoShow : boolean = false); overload;
    procedure   EnterProc(Params : array of const; ActionType : TXAMExceptionAction = ExceptRaise; DoShow : boolean = false); overload;
    procedure   LeaveProc;
    function    DoProcess : boolean;
  end;

var
  XAMDebug : TXAMDebug = nil;

//-------------------------------------------------------------------------------------------------
// XAMInitializeDebug
//-------------------------------------------------------------------------------------------------
procedure XAMInitializeDebug;
begin
  if (XAMDebug <> nil) then Exit;
  try
    XAMInitializeSystem;
    case (XAMApplication.DelphiOwner) of
      True  : XAM_DEBUG := false;
      False : case (XAMApplication.CmdLine.IndexOf('/DEBUG') <> -1) of
                  True  : XAM_DEBUG := true;
                  False : if (XAMApplication.CmdLine.IndexOf('/NODEBUG') <> -1) then XAM_DEBUG := false;
              end;
    end;
    XAMDebug := TXAMDebug.Create;
    XAMDebug.SetActive(XAM_DEBUG);
  except
    on E : Exception do
      case (ModuleIsPackage) of
        True  : ShowMessage('XAMDebugger initialisation error : ' + ENDL2 + E.Message);
        False : raise;
      end;
  end;
end;

//-------------------------------------------------------------------------------------------------
// XAMDebugger
//-------------------------------------------------------------------------------------------------
function XAMDebugger : IXAMDebug;
begin
  if (XAMDebug = nil) then XAMInitializeDebug;
  result := XAMDebug;
end;

//-------------------------------------------------------------------------------------------------
// XAMCheckIndex
//-------------------------------------------------------------------------------------------------
// PURPOSE  : Check an index.
// IN       : IndexName   -> Index var name
//            IndexValue  -> Index value
//            MinValue    -> Low bound
//            MaxValue    -> High bound
//            ExceptClass -> Exception class to raise on error, nil for ERangeError
// EXCEPT   : ERangeError or ExceptClass
//-------------------------------------------------------------------------------------------------
procedure XAMCheckIndex(IndexName : string; IndexValue, MinValue, MaxValue : integer; ExceptClass : ExceptClass);
begin
  if not (IndexValue in [MinValue..MaxValue]) then
    begin
      if (ExceptClass = nil) then ExceptClass := ERangeError;
      raise ExceptClass.CreateFmt(XAM_ERR_INVALID_INDEX, [IndexName, IndexValue, MinValue, MaxValue]);
    end;
end;

procedure XAMCheckIndex(IndexName : string; IndexValue, MinValue, MaxValue : int64; ExceptClass : ExceptClass);
begin
  if not ( (IndexValue >= MinValue) and (IndexValue <= MaxValue) ) then
    begin
      if (ExceptClass = nil) then ExceptClass := ERangeError;
      raise ExceptClass.CreateFmt(XAM_ERR_INVALID_INDEX, [IndexName, IndexValue, MinValue, MaxValue]);
    end;
end;

{
<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
                                                        TXAMMapFile
<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
}

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.Create
//-------------------------------------------------------------------------------------------------
constructor TXAMMapFile.Create;
begin
  inherited;
  FSegments := TStringList.Create;
  FAdress   := TStringList.Create;
  FLines    := TStringList.Create;
  FPresent  := false;
  FUnitName := XAM_CAPTION_UNKNOWN;
  FProcName := XAM_CAPTION_UNKNOWN;
  FLineNo   := XAM_CAPTION_UNKNOWN;
end;

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.Destroy
//-------------------------------------------------------------------------------------------------
destructor TXAMMapFile.Destroy;
begin
  FreeObject(FSegments);
  FreeObject(FAdress);
  FreeObject(FLines);
  inherited;
end;

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.FirstCallStack
//-------------------------------------------------------------------------------------------------
// PURPOSE  : Se positionne sur la premire adresse de la pile des appels (dernier call).
//-------------------------------------------------------------------------------------------------
function TXAMMapFile.FirstCallStack : boolean;
var rec : pointer;
begin
  rec := @FStackInfos;
  asm
    push ebx
    mov eax, rec
    mov ebx, ss:[ebp]   // prior_ebp
    mov [eax], ebx
    mov ebx, ss:[ebp+4] // caller_addr
    mov [eax+4], ebx
    pop ebx
  end;
  Addr   := cardinal(FStackInfos.CallerAddr);
  result := CompareText(FProcName, 'TlsLast') <> 0;
end;

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.NextCallStack
//-------------------------------------------------------------------------------------------------
// PURPOSE  : Se positionne sur l'appel suivant (remonte ainsi vers la premire procdure appele).
//-------------------------------------------------------------------------------------------------
function TXAMMapFile.NextCallStack : boolean;
var rec : pointer;
begin
  rec := @FStackInfos;
  asm
    push ebx
    push ebp
    mov eax, rec
    mov ebx, [eax]      // prior_ebp
    mov ebp, ebx        // save prior_ebp in ebp
    mov ebx, ss:[ebp]   // prior_ebp
    mov [eax], ebx
    mov ebx, ss:[ebp+4] // caller_addr
    mov [eax+4], ebx
    pop ebp
    pop ebx
  end;
  Addr   := cardinal(FStackInfos.CallerAddr);
  result := CompareText(FProcName, 'TlsLast') <> 0;
end;

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.SetFileName
//-------------------------------------------------------------------------------------------------
procedure TXAMMapFile.SetFileName(Value : TFileName);
var i, ideb, iend, i1, i2, i2b, i3 : integer;
begin
  if (FFileName = Value) then Exit;
  if (Value <> '') and (not FileExists(Value)) then raise EMapFileException.CreateFmt(ERR_MAP_FILE_NOT_FOUND, [Value]);
  FSegments.Clear;
  FAdress.Clear;
  FLines.Clear;
  FPresent := false;
  if (Value = '') then Exit;
  with (TStringList.Create) do
    try
      LoadFromFile(Value);
      i1   := 0;
      i2   := 0;
      i2b  := 0;
      i3   := 0;
      ideb := 0;
      iend := Pred(Count);
      for i := ideb to iend do if (Pos('Detailed map of segments', Strings[i]) <> 0)           then begin i1  := i; break; end;
      for i := i1   to iend do if (Pos('  Address         Publics by Name', Strings[i]) <> 0)  then begin i2  := i; break; end;
      for i := i2   to iend do if (Pos('  Address         Publics by Value', Strings[i]) <> 0) then begin i2b := i; break; end;
      for i := i2b  to iend do if (Pos('Line numbers for ', Strings[i]) <> 0)                  then begin i3  := i; break; end;
      if (i1 = 0) or (i2b = 0) or (i3 = 0) then raise EMapFileException.CreateFmt(ERR_INVALID_MAP_FILE, [Value]);
      for i := i1  to Pred(i2) do FSegments.Add(Strings[i]);
      for i := i2b to Pred(i3) do FAdress.Add(Strings[i]);
      for i := i3  to iend     do FLines.Add(Strings[i]);
      FPresent  := true;
      FFileName := Value;
    finally
      Free;
    end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.SetAddr
//-------------------------------------------------------------------------------------------------
procedure TXAMMapFile.SetAddr(Value : cardinal);
var MemInfo : TMemoryBasicInformation;
begin
  VirtualQuery(pointer(Value), MemInfo, SizeOf(MemInfo));
  Value := Value - ( cardinal(MemInfo.AllocationBase) + MemInfo.State );
  if (not FPresent) or (FAddr = Value) then Exit;
  FAddr := VAlue;
  UpdateUnitName;
  UpdateProcName;
  UpdateLineNo;
end;

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.UpdateUnitName
//-------------------------------------------------------------------------------------------------
procedure TXAMMapFile.UpdateUnitName;
var i, base, spos : integer;
begin
  FUnitName := XAM_CAPTION_UNKNOWN;
  if (not FPresent) then Exit;
  with (FSegments) do
    for i := 0 to Pred(Count) do
      begin
        try
          base := StrToInt('$' + Copy(Strings[i], MAPFILE_M_1, ADDR_LENGTH));
        except
          continue;
        end;
        if (FAddr >= base) and (FAddr < (base + StrToInt('$' + Copy(Strings[i], MAPFILE_M_2, ADDR_LENGTH)))) then
          begin
            spos := Pos('M=', Strings[i]) + 2;
            FUnitName := Trim(Copy(Strings[i], spos, Pos('ACBP=', Strings[i]) - spos - 1));
            break;
          end;
      end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.UpdateProcName
//-------------------------------------------------------------------------------------------------
procedure TXAMMapFile.UpdateProcName;
var i, val, paddr : integer;
begin
  FProcName := XAM_CAPTION_UNKNOWN;
  FProcAddr := 0;
  if (not FPresent) then Exit;
  val := 0;
  with (FAdress) do
    for i := 0 to Pred(Count) do
      begin
        try
          paddr := val;
          val   := StrToInt('$' + Copy(Strings[i], MAPFILE_M_1, ADDR_LENGTH));
        except
          continue;
        end;
        if (val = FAddr) then
          begin
            FProcName := Trim(Copy(Strings[i], MAPFILE_FCT, Length(Strings[i]) - MAPFILE_FCT + 1));
            FProcAddr := val;
            break;
          end;
        if (val > FAddr) then
          begin
            FProcName := Trim(Copy(Strings[i - 1], MAPFILE_FCT, Length(Strings[i - 1]) - MAPFILE_FCT + 1));
            FProcAddr := paddr;
            break;
          end;
      end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.UpdateLineNo
//-------------------------------------------------------------------------------------------------
procedure TXAMMapFile.UpdateLineNo;
var i, j, val, line : integer;
    b : boolean;
begin
  FLineNo := XAM_CAPTION_UNKNOWN;
  if (not FPresent) then Exit;
  line := 0;
  val  := 0;
  b    := false;
  with (FLines) do
    begin
      for i := 0 to Pred(Count) do
        if (Pos(FUnitName, Strings[i]) <> 0) then
          begin
            b := true;
            break;
          end;
      if (not b) then Exit;
      i := i + 2;
      while (b) do
        try
          for j := 0 to 3 do
            begin
              val := StrToInt('$' + Copy(Strings[i], MAPFILE_LINE_2 + MAPFILE_LINE_3 * j , ADDR_LENGTH));
              if (val > FAddr) then
                begin
                  FLineNo := IntToStr(line);
                  b := false;
                  break;
              end;
              line := StrToInt(StringReplace(Copy(Strings[i], MAPFILE_LINE_1 + MAPFILE_LINE_3 * j , 6), ' ', '', [rfReplaceAll]));
              if (val = FAddr) then
                begin
                  FLineNo := IntToStr(line);
                  b := false;
                  break;
              end;
            end;
          Inc(i);
        except
          break;
        end;
    end;
end;

{
<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
                                                        TXAMDebug
<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
}

//-------------------------------------------------------------------------------------------------
// TXAMDebug.Create
//-------------------------------------------------------------------------------------------------
constructor TXAMDebug.Create;
var s : string;
begin
  inherited;
  FProcCalls   := TCollection.Create(TCallProcInfo);
  FMapFile     := TXAMMapFile.Create;
  FObjectInfos := TXAMObjectInfos.Create;
  case (XAMApplication.DelphiOwner) and (ModuleIsPackage) of
    True  : s := XAMApplication.PathRoot + 'Projects\Bpl\pXAM.map';
    False : s := ChangeFileExt(XAMApplication.FileName, MAPFILE_EXT);
  end;
  try
    if FileExists(s) then FMapFile.FileName := s;
  except
  end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.Destroy
//-------------------------------------------------------------------------------------------------
destructor TXAMDebug.Destroy;
begin
  SetActive(false);
  FreeObject(FProcCalls);
  FreeObject(FMapFile);
  FreeObject(FObjectInfos);
  inherited;
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.QueryInterface
//-------------------------------------------------------------------------------------------------
function TXAMDebug.QueryInterface(const IID: TGUID; out Obj): Integer; stdcall;
begin
// Nothing because it's a persistant object
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug._AddRef
//-------------------------------------------------------------------------------------------------
function TXAMDebug._AddRef: Integer; stdcall;
begin
// Nothing because it's a persistant object
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug._Release
//-------------------------------------------------------------------------------------------------
function TXAMDebug._Release: Integer; stdcall;
begin
// Nothing because it's a persistent object
end;

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.GetMapFile
//-------------------------------------------------------------------------------------------------
function TXAMDebug.GetMapFile : TXAMMapFile;
begin
  result := FMapFile;
end;

//-------------------------------------------------------------------------------------------------
// TXAMMapFile.GetCallStack
//-------------------------------------------------------------------------------------------------
function TXAMDebug.GetCallStack(SkipFromLast : cardinal = 0; GetOne : boolean = false) : string;
var i : cardinal;
begin
  i      := 0;
  result := '';
  with (FMapFile) do
    if (FirstCallStack) then
      try
        repeat
          case (i = SkipFromLast) of
            True  : begin
                      if (result <> '') then result := result + ENDL;
                      result := result + UnitName + '::' + ProcName;
                      if (GetOne) then Exit;
                    end;
            False : Inc(i);
          end;
        until (not NextCallStack);
      except
      end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.IsActive
//-------------------------------------------------------------------------------------------------
function TXAMDebug.IsActive : boolean;
begin
  result := FActive;
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.SetActive
//-------------------------------------------------------------------------------------------------
procedure TXAMDebug.SetActive(Value : boolean);
begin
  if (FActive = Value) then Exit;
  case (Value) of
    True  : begin
              FOldHandler             := Application.OnException;
              Application.OnException := HandleException;
            end;
    False : Application.OnException := FOldHandler;
  end;
  FActive := Value;
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.Set/Get...Exception Events
//-------------------------------------------------------------------------------------------------
procedure TXAMDebug.SetShowException  (Value : TXAMShowExceptionEvent);   begin FShowException   := Value;  end;
procedure TXAMDebug.SetBeforeException(Value : TXAMBeforeExceptionEvent); begin FBeforeException := Value;  end;
procedure TXAMDebug.SetOnException    (Value : TXAMOnExceptionEvent);     begin FOnException     := value;  end;
procedure TXAMDebug.SetAfterException (Value : TXAMAfterExceptionEvent);  begin FAfterException  := value;  end;
function  TXAMDebug.GetShowException   : TXAMShowExceptionEvent;          begin result := FShowException;   end;
function  TXAMDebug.GetBeforeException : TXAMBeforeExceptionEvent;        begin result := FBeforeException; end;
function  TXAMDebug.GetOnException     : TXAMOnExceptionEvent;            begin result := FOnException;     end;
function  TXAMDebug.GetAfterException  : TXAMAfterExceptionEvent;         begin result := FAfterException;  end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.GetExceptionInfos
//-------------------------------------------------------------------------------------------------
function TXAMDebug.GetExceptionInfos(Sender : TObject; E : Exception) : TXAMExceptionInfos;
var i : integer;
begin
  result.EClassName  := XAM_CAPTION_UNKNOWN;
  result.EInherits   := XAM_CAPTION_UNKNOWN;
  result.EFullName   := XAM_CAPTION_UNKNOWN;
  result.EMessage    := XAM_CAPTION_UNKNOWN;
  result.EAddress    := XAM_CAPTION_UNKNOWN;
  result.AppName     := XAM_CAPTION_UNKNOWN;
  result.AppPath     := XAM_CAPTION_UNKNOWN;
  result.AppVer      := XAM_CAPTION_UNKNOWN;
  result.Date        := XAM_CAPTION_UNKNOWN;
  result.User        := XAM_CAPTION_UNKNOWN;
  result.Computer    := XAM_CAPTION_UNKNOWN;
  result.CodeUnit    := XAM_CAPTION_UNKNOWN;
  result.CodeLine    := XAM_CAPTION_UNKNOWN;
  result.CodeClass   := XAM_CAPTION_UNKNOWN;
  result.CodeProc    := XAM_CAPTION_UNKNOWN;
  result.SenderClass := XAM_CAPTION_UNKNOWN;
  result.SenderName  := XAM_CAPTION_UNKNOWN;
  result.CallStack   := XAM_CAPTION_UNKNOWN;
  if (E = nil) or (not Assigned(E)) then Exit;
  try
    FObjectInfos.Target := E;
    result.E            := E;
    result.EClassName   := FObjectInfos.TargetClassName;
    result.EInherits    := FObjectInfos.ParentClassName;
    result.EFullName    := Format('%s ( %s )', [FObjectInfos.TargetClassName, FObjectInfos.ParentClassName]);
    result.EMessage     := E.Message;
    result.EAddress     := '0x' + IntToHex(cardinal(ExceptAddr), 8);
    result.AppName      := XAMApplication.Title;
    result.AppPath      := LowerCase(XAMApplication.FileName);
    if (XAMApplication.Version <> '') then result.AppVer := XAMApplication.Version;
    result.Date         := DateTimeToStr(Now);
    result.User         := XAMWindows.User;
    result.Computer     := XAMWindows.Computer;
    with (XAMDebug.FMapFile) do
      begin
        Addr            := cardinal(ExceptAddr);
        result.CodeUnit := UnitName;
        result.CodeLine := LineNo;
        i := Pos('.', ProcName);
        case (i = 0) of
          True  : begin
                    result.CodeClass := XAM_CAPTION_NA;
                    result.CodeProc  := ProcName;
                  end;
          False : begin
                    result.CodeClass := Copy(ProcName, 1, i - 1);
                    result.CodeProc  := Copy(ProcName, i + 1, Length(ProcName));
                  end;
        end;
      end;
    case Assigned(Sender) of
      True  : begin
                FObjectInfos.Target  := Sender;
                result.SenderClass  := FObjectInfos.Inherits;
                result.SenderName   := FObjectInfos.ComponentName;
              end;
      False : begin
                result.SenderClass  := XAM_CAPTION_NA;
                result.SenderName   := XAM_CAPTION_NA;
              end;
    end;
    result.CallStack := XAMDebug.GetCallStack;
  except
  end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.HandleException
//-------------------------------------------------------------------------------------------------
// PURPOSE  : Main exception manager.
//-------------------------------------------------------------------------------------------------
procedure TXAMDebug.HandleException(Sender : TObject; E : Exception);
begin
  if (E is EAbort) then Exit;
  ProcessException(Sender, E, true);
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.ProcessException
//-------------------------------------------------------------------------------------------------
procedure TXAMDebug.ProcessException(Sender : TObject; E : Exception; DoShow : boolean);
var b : boolean;
    s : string;
begin
  b := true;
  try
    try
      s := StringOfChar(' ', FProcCallsMargin);
      with XAMDebugger.GetExceptionInfos(Sender, E) do
        XAMApplication.LogFile.Write(logExcept,
                                     s + 'Type         : ' + EFullName   + ENDL +
                                     s + 'Unit         : ' + CodeUnit    + ENDL +
                                     s + 'Line         : ' + CodeLine    + ENDL +
                                     s + 'Class        : ' + CodeClass   + ENDL +
                                     s + 'Function     : ' + CodeProc    + ENDL +
                                     s + 'Address      : ' + EAddress    + ENDL +
                                     s + 'Sender Class : ' + SenderClass + ENDL +
                                     s + 'Sender Name  : ' + SenderName  + ENDL +
                                     s + 'Message      : ' + FormatText(EMessage, 0, 15 + Length(s)) + ENDL +
                                     s + 'Call Stack   : ' + FormatText(CallStack, 0, 15 + Length(s)));

    except
    end;
    if Assigned(FBeforeException) then
      try
        FBeforeException(Sender, E, b);
      except
        on E : Exception do
          if (DoShow) then XAMApplication.ShowMessageLog('Error on BeforeException :' + ENDL2 + E.Message);
      end;
    if (b) and (DoShow) then ShowException(Sender, E);
    if Assigned(FAfterException) then
      try
        FAfterException(Sender, E, b);
      except
        on E : Exception do
          if (DoSHow) then XAMApplication.ShowMessageLog('Error on AfterException :' + ENDL2 + E.Message);
      end;
  except
    on E : Exception do
      begin
        s := 'Error on processing Exception : ' + ENDL2;
        case (E <> nil) and Assigned(E) of
          True  : s := s + E.Message + ENDL2;
          False : s := s + '     [NULL Object]' + ENDL2;
        end;
        s := s + '(' + E.Message + ')';
        XAMApplication.ShowMessageLog(s);
      end;
  end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.ShowException
//-------------------------------------------------------------------------------------------------
procedure TXAMDebug.ShowException(Sender : TObject; E : Exception);
var s : string;
begin
  if (E is EAbort) then Exit;
  if Assigned(FOnException) then
    try
      FOnException(Sender, E);
    except
      on E : Exception do XAMApplication.ShowMessageLog('Error on OnException :' + ENDL2 + E.Message);
    end;
  if Assigned(Application) and (csDestroying in Application.ComponentState) then MessageAbort(E.Message);
  try
    case Assigned(FShowException) of
      True  : FShowException(Sender, E);
      False : TfmXAMException.Execute(Sender, E);
    end;
  except
    on E : Exception do
      begin
        s := 'Error on displaying Exception :' + ENDL2;
        case (E <> nil) and Assigned(E) of
          True  : s := s + E.Message + ENDL2;
          False : s := s + '     [NULL Object]' + ENDL2;
        end;
        s := s + '(' + E.Message + ')';
        XAMApplication.ShowMessageLog(s);
      end;
  end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.EnterProc
//-------------------------------------------------------------------------------------------------
// PURPOSE  : Declare an enter proc.
//-------------------------------------------------------------------------------------------------
procedure TXAMDebug.EnterProc(ActionType : TXAMExceptionAction; DoShow : boolean);
begin
  try
    with (FProcCalls.Add as TCallProcInfo) do
      begin
        Caption := GetCallStack(2, true);
        Action  := ActionType;
        Show    := DoShow;
        XAMApplication.LogFile.Write(logEnter, StringOfChar(' ', FProcCallsMargin) + Caption);
        Inc(FProcCallsMargin, 2);
      end;
  except
  end;
end;

procedure TXAMDebug.EnterProc(Params : array of const; ActionType : TXAMExceptionAction; DoShow : boolean);
var s : string;
    i : integer;
begin
  try
    with (FProcCalls.Add as TCallProcInfo) do
      begin
        Caption := GetCallStack(2, true);
        s       := Caption;
        try
          if (Length(Params) > 0) then
            begin
              s := s + ' ( ';
              for i := Low(Params) to High(Params) do s := s + VarRecToText(Params[i]) + ' , ';
              Delete(s, Length(s) - 2, 3);
              s := s + ' )';
            end;
        except
        end;
        XAMApplication.LogFile.Write(logEnter, StringOfChar(' ', FProcCallsMargin) + s);
        Inc(FProcCallsMargin, 2);
      end;
  except
  end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMDebug.LeaveProc
//-------------------------------------------------------------------------------------------------
// PURPOSE  : Declare an exit proc.
//-------------------------------------------------------------------------------------------------
procedure TXAMDebug.LeaveProc;
var i : integer;
begin
  try
    i := Pred(FProcCalls.Count);
    with (FProcCalls.Items[i] as TCallProcInfo) do
      begin
        Dec(FProcCallsMargin, 2);
        XAMApplication.LogFile.Write(logLeave, StringOfChar(' ', FProcCallsMargin) + Caption);
      end;
    FProcCalls.Delete(i);
  except
  end;
end;

//-------------------------------------------------------------------------------------------------
// DoProcess
//-------------------------------------------------------------------------------------------------
// PURPOSE  : handle an exception raised in a Enter/Leave block.
//            Return true if the exception must be raise, false if not.
//-------------------------------------------------------------------------------------------------
function TXAMDebug.DoProcess : boolean;
begin
  result := false;
  if (ExceptObject.ClassType = EAbort) then
    begin
      if (FProcCalls.Count > 0) then LeaveProc;
      raise EAbort.Create((ExceptObject as EAbort).Message);
    end;
  case (FProcCalls.Count = 0) of
    True  : begin
              ProcessException(self, ExceptObject as Exception, false);
              result := true;
            end;
    False : with (FProcCalls.Items[ Pred(FProcCalls.Count) ] as TCallProcInfo) do
              begin
                ProcessException(self, ExceptObject as Exception, Show);
                case (Action) of
                  ExceptMask  : LeaveProc;
                  ExceptRaise : begin
                                  LeaveProc;
                                  result := true;
                                end;
                  ExceptStop  : begin
                                  LeaveProc;
                                  raise EAbort.Create('');
                                end;
                end;
              end;
  end;
end;

{
#######################################################################################################################
                                                    INITIALIZATION
#######################################################################################################################
}

initialization
  try
    XAMInitializeDebug;
  except
    if (not ModuleIsPackage) then raise;
  end;

end.

