{
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
PLATFORM  : Windows 9x/NT/XP - Delphi 5,6,7
PROJECT   : eXpress Application Manager
FILE      : U_XAM_LogFile.pas
PURPOSE   : Log file manager.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
LICENSE   : LGPL
AUTHOR    : Copyright 2003 Olivier Rogier 
WEBSITE   : http://www.ordisoftware.info
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CREATION  : v 1.0.0 - 01/2003
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
}

unit U_XAM_LogFile;

{
#######################################################################################################################
                                                       INTERFACE
#######################################################################################################################
}

interface

uses
  Classes, F_XAM_Log;

type
  TXAMLogEvent = ( logSystem, logInfo, logError, logExcept, logEnter, logLeave );

  TXAMLogFile = class
  private
    FLines       : TStringList;       // Logfile content
    FFileName    : string;            // Log filename
    FEnabled     : boolean;           // True if trace enabled
    FSize        : integer;           // Max logfile size for purge
    FCacheSize   : integer;           // Cache size in inscriptions count (one write, not lines count)
    FCachePos    : integer;           // Position in cache
    FForm        : TfmXAMLog;
  private
    procedure SetEnabled(AValue : boolean);
  public
    constructor Create;
    destructor  Destroy; override;
    procedure   Show;
    procedure   Clear;
    procedure   Flush;
    procedure   Start;
    procedure   Stop;
    procedure   Write(AType : TXAMLogEvent; AText : string; DoFlush : boolean = false);
  public
    property Lines       : TStringList      read FLines;
    property FileName    : string           read FFileName    write FFileName;
    property Enabled     : boolean          read FEnabled     write SetEnabled;
    property Size        : integer          read FSize        write FSize;
    property CacheSize   : integer          read FCacheSize   write FCacheSize;
  end;

{
#######################################################################################################################
                                                     IMPLEMENTATION
#######################################################################################################################
}

implementation

uses
  SysUtils, Dialogs, Forms, U_XAM_Global, U_XAM_System;

const
  LOG_EVENT_NAME : array[TXAMLogEvent] of string
                 = ( '', 'INFO  ', 'ERROR ', 'EXCEPT', 'ENTER ', 'LEAVE ' );

//-------------------------------------------------------------------------------------------------
// TXAMLogFile.Create
//-------------------------------------------------------------------------------------------------
constructor TXAMLogFile.Create;
begin
  FLines     := TStringList.Create;
  FForm      := TfmXAMLog.Create(nil);
  FSize      := 5000;
  FCacheSize := 100;
end;

//-------------------------------------------------------------------------------------------------
// TXAMLogFile.Create
//-------------------------------------------------------------------------------------------------
destructor TXAMLogFile.Destroy;
begin
  Stop;
  FreeObject(FForm);
  FreeObject(FLines);
  inherited;
end;

//-------------------------------------------------------------------------------------------------
// TXAMLogFile.Show
//-------------------------------------------------------------------------------------------------
procedure TXAMLogFile.Show;
begin
  if (not FEnabled) then Exit;
  FForm.Execute(FLines);
end;

//-------------------------------------------------------------------------------------------------
// TXAMLogFile.SetEnabled
//-------------------------------------------------------------------------------------------------
procedure TXAMLogFile.SetEnabled(AValue: boolean);
begin
  if (FEnabled = AValue) then Exit;
  FEnabled := AValue;
  case (AValue) of
    True  : Start;
    False : Stop;
  end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMLogFile.Clear
//-------------------------------------------------------------------------------------------------
procedure TXAMLogFile.Clear;
begin
  if (not FEnabled) then Exit;
  FLines.Clear;
  Flush;
end;

//-------------------------------------------------------------------------------------------------
// TXAMLogFile.Flush
//-------------------------------------------------------------------------------------------------
procedure TXAMLogFile.Flush;
begin
  if (not FEnabled) then Exit;
  FLines.SaveToFile(FFileName);
end;

//-------------------------------------------------------------------------------------------------
// TXAMLogFile.Start
//-------------------------------------------------------------------------------------------------
procedure TXAMLogFile.Start;
begin
  if (not FEnabled) then Exit;
  try
    case FileExists(FFileName) of
      True  : FLines.LoadFromFile(FFileName);
      False : CreateEmptyFile(FFileName);
    end;
    Write(logSystem, '-----------------------------------------------------------------------------------------');
    Write(logSystem, 'START LOGGING : ' + XAMApplication.Title + ' at ' + DateTimeToStr(Now));
    Write(logSystem, '-----------------------------------------------------------------------------------------');
    Write(logSystem, 'PROCESS  : ' + XAMApplication.FileName);
    Write(logSystem, 'USER     : ' + XAMWindows.User);
    Write(logSystem, 'COMPUTER : ' + XAMWindows.Computer);
    Write(logSystem, 'SYSTEM   : ' + XAMWindows.Version);
    Write(logSystem, '-----------------------------------------------------------------------------------------');
    Write(logSystem, '');
  except
    on E : Exception do
      begin
        FEnabled := false;
        ShowMessage('Error on starting logging ' + FFileName + ENDL +
                    'Trace is disabled...' + ENDL2 +
                    E.Message);
      end;
  end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMLogFile.Stop
//-------------------------------------------------------------------------------------------------
procedure TXAMLogFile.Stop;
begin
  if (not FEnabled) then Exit;
  try
    try
      Write(logSystem, '');
      Write(logSystem, '-----------------------------------------------------------------------------------------');
      Write(logSystem, 'STOP LOGGING at ' + DateTimeToStr(Now));
      Write(logSystem, '-----------------------------------------------------------------------------------------');
      Write(logSystem, '');
    finally
      Flush;
    end;
  except
  end;
end;

//-------------------------------------------------------------------------------------------------
// TXAMLogFile.Write
//-------------------------------------------------------------------------------------------------
procedure TXAMLogFile.Write(AType: TXAMLogEvent; AText: string; DoFlush: boolean);
Var s : string;
begin
  try
    if (not FEnabled) then Exit;
    if not FileExists(FFileName) then CreateEmptyFile(FFileName);
    s := '# ';
    if (AType <> logSystem) then s := s + DateTimeToStr(Now) + ' [ ' + LOG_EVENT_NAME[AType] + ' ] ';
    FLines.Add(FormatText(s + AText, 0, Length(s), ' '));
    Inc(FCachePos);
    while (FLines.Count > FSize) do FLines.Delete(0);
    if (DoFlush) or (FCachePos >= FCacheSize) then
      begin
        Flush;
        FCachePos := 0;
      end;
  except
  end;
end;

end.

